from nbodykit.lab import *
from nbodykit import setup_logging, style
from nbodykit.source.catalog import BinaryCatalog
import numpy as np
import matplotlib.pyplot as plt
from numpy import linalg as la
from scipy.interpolate import InterpolatedUnivariateSpline
from nbodykit import CurrentMPIComm
import os
import sys



#Add RSD to catalog using a specified fixed LOS
def add_RSD_fixedLOS(data,f,LOS):
    product = (np.asarray(data['VelocityOffset']*LOS)).sum(1)
    print("shift is %f " %(product[0]))
    data['Position'] = data['Position'] + LOS*product[:,np.newaxis]
    return data

# Write out the poles ide
def write_files(poles, dirPath,dataSeed):
    comm = CurrentMPIComm.get()
    if(comm.rank !=0):
        return
        
    modes = np.asarray(poles['modes'])
    kvals = np.asarray(poles['k'])
    if dataSeed == 1:
        kval_filename = os.path.join(dirPath,'kvals_seed_%d' %dataSeed)
        np.savetxt(kval_filename, kvals, fmt='%.7e')
    if dataSeed == 1:
        mode_filename = os.path.join(dirPath,'modes.txt')
        np.savetxt(mode_filename,modes,fmt='%.7e')
    power_array_real = []
    power_array_im = []

    for ell in [0,1,2,3,4]:
        Pre = np.asarray(poles['power_%d' %ell].real)
        Pim = np.asarray(poles['power_%d' %ell].imag)
        power_array_im.append(Pim)
        power_array_real.append(Pre)

    fNameIm = os.path.join(dirPath,'power_imag_seed_%d.txt' %dataSeed)
    fNameRe = os.path.join(dirPath,'power_real_seed_%d.txt' %dataSeed)
    np.savetxt(fNameIm,power_array_im,fmt='%.7e')
    np.savetxt(fNameRe,power_array_real,fmt='%.7e')

# Cut catalog to make sure all positions fall within the box. This is relevant when RSD is added. 
def cut_catalog(data,L):
    select = (-L/2. <= data['Position'][:,0])&(data['Position'][:,0]<= L/2.)
    data = data[select]
    select = (-L/2. <= data['Position'][:,1])&(data['Position'][:,1]<= L/2.)
    data = data[select]
    select = (-L/2. + shift <= data['Position'][:,2])&(data['Position'][:,2]<= L/2.)
    data = data[select]
    return data

#Make FKP catalog and call ConvolvedFFT algorithm. 

#Make a mesh and call FFTPower
def call_FFTPower(data,nmesh,L,LOS):
    real_mesh = data.to_mesh(compensated=True, BoxSize = L, Nmesh=nmesh, resampler='tsc', position='Position')
    #compute the 2d P(k,mu) power, with 5 mu bins
    r = FFTPower(real_mesh, mode='2d', poles=[0,1,2,3,4],los=LOS)
    print(r.poles.attrs['shotnoise'])
    r.poles['power_0'].real = r.poles['power_0'].real - r.poles.attrs['shotnoise']
    return r

#Take the mean value of a specified number of simulations. 
def average_poles(start, stop, importPath, exportPath, do_error):
    comm = CurrentMPIComm.get()
    if(comm.rank !=0):
        return
    #How do we know what rank it is???
    print("rank is %d" %(comm.rank))
    filename = os.path.join(importPath,'kvals_seed_1')
    kvals = np.loadtxt(filename)
    fn = os.path.join(exportPath,'kvals.txt')
    np.savetxt(fn,kvals,fmt='%.7e')

    real_power = []
    imag_power = []

    newRange = np.arange(start,stop,1)

    for seed in newRange: 
        Real_fn = os.path.join(importPath,'power_real_seed_%d.txt' %seed)
        Imag_fn = os.path.join(importPath,'power_imag_seed_%d.txt' %seed)
        print(Real_fn)
        real_power.append(np.loadtxt(Real_fn))
        imag_power.append(np.loadtxt(Imag_fn))

    real_power = np.asarray(real_power)
    imag_power = np.asarray(imag_power)

    average_real = np.mean(real_power,axis=0)
    average_imag = np.mean(imag_power,axis=0)

    for ell in [0,1,2,3,4]:
        if ell%2==0:
            P = average_real[ell,:]
            PReal = real_power[:,ell,:]
            fn = os.path.join(exportPath,'power_%d.txt' %ell)
            np.savetxt(fn,P,fmt='%.7e')

            if(do_error == True):
                cov = np.cov(PReal,rowvar=False)
                error = np.sqrt(np.diagonal(cov))
                fn = os.path.join(exportPath,'error_%d.txt' %ell)
                np.savetxt(fn,error,fmt='%.7e')
            
            

        elif ell%2==1:
            P = average_imag[ell,:]
            PImag = imag_power[:,ell,:]
            fn = os.path.join(exportPath,'power_%d.txt' %ell)
            np.savetxt(fn,P,fmt='%.7e')

            if(do_error == True):
                cov = np.cov(PImag,rowvar=False)
                error = np.sqrt(np.diagonal(cov))
                fn = os.path.join(exportPath,'error_%d.txt' %ell)
                np.savetxt(fn,error,fmt='%.7e')

# Kaiser functions for plotting

def kaiser_pkmu(k, mu,b1,f,Plin):
    beta = f / b1
    return (1 + beta*mu**2)**2 * b1**2 * Plin(k)

def kaiser_p0(k,b1,f,Plin):
    beta = f / b1
    fac = 1 + (2./3.)*beta + (1./5.)*beta*beta
    return fac*Plin(k)*b1*b1

def kaiser_p2(k,b1,f,Plin):
    beta = f / b1
    fac = (4./3.)*beta + (4./7.)*beta*beta
    return fac*b1*b1*Plin(k)
    
def kaiser_p4(k,b1,f,Plin):
    beta = f / b1
    fac = (8./35.)*beta*beta
    return fac*b1*b1*Plin(k)


# Make a plot of the poles. 

def make_plot(r,Plin, exportPath, b1, f,do_MatterPow, do_Kaiser,plotTitle,plotFN, do_error):
    comm = CurrentMPIComm.get()
    if(comm.rank !=0):
        return

    colors = ["b","r","orange","darkorchid","c"]
    fn_k = os.path.join(exportPath,'kvals.txt')
    kvals = np.loadtxt(fn_k)

    MatterPow = kaiser_pkmu(kvals, 0,b1,f,Plin)

    for ell in [0,1,2,3,4]:
        if ell%2==0:
            fn_pow = os.path.join(exportPath,'power_%d.txt' %ell)
            P = np.loadtxt(fn_pow)
            if(do_error == True):  
                fn_err = os.path.join(exportPath,'error_%d.txt' %ell)
                error = np.loadtxt(fn_err)
                plt.errorbar(kvals, P,yerr=error,ls = 'none',color=colors[ell])
            label = r'$\ell=%d$' % (ell)
            plt.plot(kvals, P,'.',color=colors[ell],label = label)
            

        elif ell%2==1:
            fn_pow = os.path.join(exportPath,'power_%d.txt' %ell)
            P = np.loadtxt(fn_pow)
            if(do_error == True):  
                fn_err = os.path.join(exportPath,'error_%d.txt' %ell)
                error = np.loadtxt(fn_err)
                plt.errorbar(kvals, P/(b1*b1),yerr=error,ls = 'none',color=colors[ell])
            label = r'$\ell=%d$' % (ell)
            plt.plot(kvals, P,'.',color=colors[ell],label = label)

    if(do_Kaiser == True):
        plt.plot(kvals,kaiser_p0(kvals,b1,f,Plin),'--',color='black',label = 'Kaiser')
        plt.plot(kvals,kaiser_p2(kvals,b1,f,Plin),'--',color='black')
        plt.plot(kvals,kaiser_p4(kvals,b1,f,Plin),'--',color='black')
    
    if(do_MatterPow == True):
        plt.plot(kvals,kaiser_pkmu(kvals, 0,b1,f,Plin),color='blue',label = 'measured')

    plt.xlabel(r"$k$ [$h \ \mathrm{Mpc}^{-1}$]")
    plt.ylabel(r"k $P_\ell$ [$h^{-2} \mathrm{Mpc}^2$]")
    #plt.ylim(-1e5,1e5)
    plt.xscale("log")
    plt.title(plotTitle)
    plt.legend(loc=0)
    figname = os.path.join(exportPath, plotFN)
    print(figname)
    plt.savefig(figname, dpi=None, facecolor='w', edgecolor='w',
        orientation='portrait', papertype=None, format='png',
        transparent=False, bbox_inches=None, pad_inches=0.1) 


# define parameters 

Catalog_path = "/scratch/07492/iss23/Catalogs/Final_SPHEREx/" # only needed if not using Nbodykit lognormals
import_dirpath = ""
export_dirpath = ""
n = 0.0004
shift = 0 # set to 100000 to approximate fixed LOS for Yamamoto
L = 2640
b1 = 1.5
nmesh = 256
LOS = [0,0,1]
startSeed = 1
endSeed = 51
SeedRange = np.arange(startSeed,endSeed)
PlotTitle = "FFTPower Debugging, Nbodykit lognorms"
PlotFn = "FFTPower_Nbodykit.png"

#Flags
do_averaging = True #Average over multiple simulations
do_Error = True # calculate error bars; set to false if only doing one simulation. 
do_Plot = True #Plot the multipoles
do_MatterPow = False #Plot matter power spectrum
do_Kaiser = True # Plot Kaiser multipoles
do_Julia = False # Read in Julia lognormal catalogs instead of using Nbodykit's (these don't have FoG)

for dataSeed in SeedRange:

    cosmo = cosmology.Cosmology(h=0.676).match(Omega0_m=0.31)
    redshift = 0.3
    f = cosmo.scale_independent_growth_rate(redshift)
    print("f is %f" %f)
    Plin = cosmology.LinearPower(cosmo, redshift, transfer='EisensteinHu')
    # Plin currently does not match Julia catalogs and will not until I change it

    if(do_Julia == True):
        fname = os.path.join(Catalog_path,'lognorm_pos_%d.txt' %dataSeed)
        data = BinaryCatalog(fname, [('Position', ('f4', 3)), ('VelocityOffset', ('f4', 3))])
        # Need to covert velocity from Mpc/h to *f
        data['VelocityOffset'] = f*data['VelocityOffset'] # units of Mpc/ h
        data['NZ']= n

    else:
        data = LogNormalCatalog(Plin=Plin, nbar=n, BoxSize=L, Nmesh=nmesh, bias=b1, seed=dataSeed)
        data['Position'] += [-L/2.,-L/2.,-L/2.]

    
    
    print("length of data is %d" %(len(data['Position'])))
    #print("%f %f %f" %(data['Position'][0,0],data['Position'][0,1], data['Position'][0,2]))

    
    data = add_RSD_fixedLOS(data,f,LOS)

    data = cut_catalog(data,L)
    print("length of data is %d after cuts" %(len(data['Position'])))

    
    r = call_FFTPower(data,nmesh,L,LOS)

    poles = r.poles
    print(poles.attrs['shotnoise'])

    write_files(poles,import_dirpath,dataSeed)


if(do_averaging == True):
    average_poles(startSeed,endSeed,import_dirpath, export_dirpath, do_Error)

if(do_Plot == True):
    make_plot(r,Plin,export_dirpath,b1,f,do_MatterPow,do_Kaiser,PlotTitle,PlotFn, do_Error)